from typing import Dict

from sara_sdk.common.session import Session
from ...utils.rest import retrieve as _retrieve, list as _list, update as _update, delete as _delete, create as _create

RESOURCE = "iam/users"


def list(session: Session = None, **filters):
    """
    List a array of users

    Args:
      session (Session): Used only if want to use a different session instead default
      filters (Any): filters to pass to filter the list of users

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> list(page=1,page_size=10,name="user name")
    """
    result = _list(resource=RESOURCE, session=session, **filters)
    return result


def retrieve(uuid: str, session: Session = None):
    """
    Retrieve a user by uuid

    Args:
      uuid (string): user uuid to retrieve
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> retrieve("f8b85a7a-4540-4d46-a2ed-00e6134ee84a")
    """
    result = _retrieve(RESOURCE, id=uuid, session=session)
    return result


def update(uuid: str, model: Dict, session: Session = None):
    """
    Update a user by passing uuid and an model (Data to update)

    Args:
      uuid (string): user uuid to retrieve
      model (Dict): A dictionary with the data the will be updated on user
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> update("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", { "name": "new name" })
    """
    result = _update(RESOURCE, id=uuid, payload=model, session=session)
    return result


def create(model: Dict, session: Session = None):
    """
    Create a user by passing an model (Data)

    Args:
      model (Dict): A dictionary with the data the will be used to create an user
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json
    """
    result = _create(RESOURCE, payload=model, session=session)
    return result


def delete(uuid: str, session: Session = None):
    """
    Delete a user by passing uuid

    Args:
      uuid (string): user uuid to delete
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> delete("f8b85a7a-4540-4d46-a2ed-00e6134ee84a")
    """
    result = _delete(RESOURCE, id=uuid, session=session)
    return result


def me(session: Session = None):
    """
    Get the data from the user logged

    Args:
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> me()
    """
    result = _retrieve(RESOURCE, id="me", session=session)
    return result


def verifyUserByEmail(email: str, session: Session = None):
    """
    Verify user by email

    Args:
      email (string): email to validate
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> verifyUserByEmail("email@dominio.com")
    """
    result = _retrieve(RESOURCE, id="verifyUserByEmail",
                       session=session, email=email)
    return result



from base64 import b64encode
from requests import post
from time import time
from sys import version_info as python_version
from ..error import InternalServerError, UnknownError, Error
from ..client.requests import Response
import os
import sara_sdk


class Session:
    """
    Session model responsable to maintain data access, you can use the DEFAULT_SESSION by just
    using the sara_sdk.auth() but can create other sessions using this class and passing on
    others functions by the param session

    Args:
        access_key (string): Access Key used to authenticate the user on Sara API
        secret_access_key (string): Secret Key used to authenticate the user on Sara API
    """

    def __init__(self, access_key, secret_access_key, scope=None):
        self.access_key = access_key
        self.secret_access_key = secret_access_key
        self.scope = scope
        self.access_token = ""
        self.attemps = 0

    def auth(self):
        """
        Authenticate the session using the oauth2.0 on aws cognito

        Returns:
            Response: A object of Response with status and content data

        Examples:
            >>> result = Session.auth();
        """
        url = sara_sdk.AUTH_URL + \
            "?client_id={client_id}".format(client_id=self.access_key)
        if self.scope is not None:
            body = {
                "grant_type": "client_credentials",
                "scope": self.scope
            }
        else:
            body = {
                "grant_type": "client_credentials",
            }

        agent = "Python-{major}.{minor}.{micro}-Sara-SDK-{sdk_version}".format(
            major=python_version.major,
            minor=python_version.minor,
            micro=python_version.micro,
            sdk_version=sara_sdk.__version__,
        )

        auth = "{access}:{secret}".format(
            access=self.access_key, secret=self.secret_access_key)

        try:
            request = post(url=url,
                           data=body,
                           headers={
                               "Content-Type": "application/x-www-form-urlencoded",
                               "User-Agent": agent,
                               "Accept-Language": "en-US",
                               "Authorization": "Basic {}".format(b64encode(auth.encode()).decode())
                           },
                           timeout=sara_sdk.__timeout__)
        except Exception as exception:
            raise UnknownError("{}: {}".format(
                exception.__class_.__name__, str(exception)))

        response = Response(status=request.status_code,
                            content=request.content)

        if response.content is not None:
            self.access_token = response.json()["access_token"]
            self.expires_in = response.json()["expires_in"] + time()
            self.token_type = response.json()["token_type"]

        if response.status == 500:
            raise InternalServerError()
        if response.status == 400:
            raise Error(response.status, response.json()["detail"])
        if response.status != 200:
            raise UnknownError(response.content)

        return response



from typing import Dict

from sara_sdk.common.session import Session
from ...utils.rest import retrieve as _retrieve, list as _list, update as _update, delete as _delete, create as _create

RESOURCE = "iam/apps"


def list(session: Session = None):
    """
    List a array of apps

    Args:
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> list()
    """
    result = _list(RESOURCE, session=session)
    return result


def retrieve(id: str, session: Session = None):
    """
    Retrieve an app by id

    Args:
      id (string): app uuid to retrieve
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> retrieve("721o83iusldua3c2fls61p9i32")
    """
    result = _retrieve(RESOURCE, id=id, session=session)
    return result


def create(model: Dict, session: Session = None):
    """
    Create a app by passing an model (Data)

    Args:
      model (Dict): A dictionary with the data the will be used to create an app
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json
    """
    result = _create(RESOURCE, payload=model, session=session)
    return result


def delete(id: str, session: Session = None):
    """
    Delete a app by passing id

    Args:
      id (string): app id to delete
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> delete("721o83iusldua3c2fls61p9i32")
    """
    result = _delete(RESOURCE, id=id, session=session)
    return result



from requests import get, post, delete as _delete, patch
from ..client.requests import fetch


def retrieve(resource, id, session=None, **kwargs):
    """
    Retrieve function to request a GET passing id to the API

    Args:
        resource (string): the route to access on the api
        id (string): uuid to the resource
        session (Session): session where to get the authorization (only needed if doens't want
        to use the DEFAULT_SESSION)
        kwargs (any): will be used as query to the route

    Returns:
        Json: The json of the content of the response sended by the api

    Example:
        >>> retrieve(resource="iam/robots", "09594aae-7e88-4c8b-b4e5-095c6e785509")
    """
    path = "{endpoint}/{id}".format(endpoint=resource, id=id)
    json = fetch(method=get, path=path, query=kwargs, session=session).json()
    return json


def list_paginated(resource, session=None, version="v1", **kwargs):
    """
    List function to request a GET to receive a list of objects

    Args:
        resource (string): the route to access on the api
        session (Session): session where to get the authorization (only needed if doens't want
        to use the DEFAULT_SESSION)
        kwargs (any): will be used as query to the route

    Returns:
        Json: The json of the content results of the response sended by the api

    Example:
        >>> list(resource="iam/robots")
    """
    page = kwargs.get("page", 1)

    while True:
        json = fetch(method=get, path=resource, query={
                     **kwargs, "page": page}, session=session, version=version).json()
        yield json.get("results", [])
        if json.get("next") is None:
            break
        page += 1


def list(resource, session=None, version="v1", **kwargs):
    """
    List function to request a GET to receive a list of objects

    Args:
        resource (string): the route to access on the api
        session (Session): session where to get the authorization (only needed if doens't want
        to use the DEFAULT_SESSION)
        kwargs (any): will be used as query to the route

    Returns:
        Json: The json of the content of the response sended by the api

    Example:
        >>> list(resource="iam/robots")
    """
    json = fetch(method=get, path=resource,
                 query=kwargs, session=session, version=version).json()
    return json


def create(resource, payload, session=None, version="v1", **kwargs):
    """
    Create function to request a POST to create a new instance of resource

    Args:
        resource (string): the route to access on the api
        payload (Dict): A dict with data to use to create the new resource
        session (Session): session where to get the authorization (only needed if doens't want
        to use the DEFAULT_SESSION)
        kwargs (any): will be used as query to the route

    Returns:
        Json: The json of the content of the response sended by the api

    Example:
        >>> create(resource="iam/robots", **kwargs)
    """
    json = fetch(method=post, path=resource, session=session,
                 query=kwargs, payload=payload, version=version).json()
    return json


def delete(resource, id, session=None, version="v1"):
    """
    Delete function to request a DELETE passing id to the API

    Args:
        resource (string): the route to access on the api
        id (string): uuid to the resource
        session (Session): session where to get the authorization (only needed if doens't want
        to use the DEFAULT_SESSION)
        kwargs (any): will be used as query to the route

    Returns:
        Json: The json of the content of the response sended by the api

    Example:
        >>> delete(resource="iam/robots", "09594aae-7e88-4c8b-b4e5-095c6e785509")
    """
    path = "{endpoint}/{id}".format(endpoint=resource, id=id)
    json = fetch(method=_delete, path=path,
                 session=session, version=version).json()
    return json


def update(resource, id, payload, session=None, version="v1"):
    """
    Delete function to request a DELETE passing id to the API

    Args:
        resource (string): the route to access on the api
        id (string): uuid to the resource
        session (Session): session where to get the authorization (only needed if doens't want
        to use the DEFAULT_SESSION)
        payload (Dict): A dict with data to use to update the resource

    Returns:
        Json: The json of the content of the response sended by the api

    Example:
        >>> delete(resource="iam/robots", "09594aae-7e88-4c8b-b4e5-095c6e785509", **payload)
    """
    path = "{endpoint}/{id}".format(endpoint=resource, id=id)
    json = fetch(method=patch, path=path,
                 payload=payload, session=session, version=version).json()
    return json


def attach(resource, type, this, that, session=None, version="v1"):
    """
    Attach function to attach something (this) to other entity (that)

    Args:
        resource (string): the route to access on the api
        type (string): type of entity that is going to be attach
        this (string): an uuid from the thing that is going to be attach to the entity
        that (string): uuid from the entity
        session (Session): session where to get the authorization (only needed if doens't want
        to use the DEFAULT_SESSION)

    Returns:
        Json: The json of the content of the response sended by the api

    Example:
        >>> attach("iam/groups", "user", "9871a24a-89ff-4ba1-a350-458244e8244d", "00ea03f1-5b2e-4787-8aa6-745961c6d506")
    """
    body = {
        type: this
    }

    # Workaround to userGroup route
    if type == "user" and resource == "iam/groups":
        type = "UserGroup"

    if type == "actions" and resource == "iam/policies":
        type = "Permissions"

    path = "{endpoint}/{that}/attach{type}".format(
        endpoint=resource, that=that, type=type.capitalize())

    if resource == "iam/clients":
        body = {
            type: this,
            'client': that
        }
        path = "{endpoint}/attach{type}".format(
            endpoint=resource, type=type.capitalize())

    json = fetch(method=post, path=path, payload=body,
                 session=session, version=version).json()
    return json


def detach(resource, type, this, that, session=None, version="v1"):
    """
    Attach function to detach something (this) from other entity (that)

    Args:
        resource (string): the route to access on the api
        type (string): type of entity that is going to be detach
        this (string): an uuid from the thing that is going to be detach from the entity
        that (string): uuid from the entity
        session (Session): session where to get the authorization (only needed if doens't want
        to use the DEFAULT_SESSION)

    Returns:
        Json: The json of the content of the response sended by the api

    Example:
        >>> detach("iam/groups", "user", "9871a24a-89ff-4ba1-a350-458244e8244d", "00ea03f1-5b2e-4787-8aa6-745961c6d506")
    """

    body = {
        type: this
    }

    # Workaround to userGroup route
    if type == "user" and resource == "iam/groups":
        type = "UserGroup"

    if type == "actions" and resource == "iam/policies":
        type = "Action"

    path = "{endpoint}/{that}/attach{type}".format(
        endpoint=resource, that=that, type=type.capitalize())

    if resource == "iam/clients":
        body = {
            type: this,
            'client': that
        }
        path = "{endpoint}/attach{type}".format(
            endpoint=resource, type=type.capitalize())

    json = fetch(method=_delete, path=path,
                 payload=body, session=session, version="v1").json()
    return json



from typing import Dict

from sara_sdk.common.session import Session
from ...utils.rest import retrieve as _retrieve, list as _list, update as _update, delete as _delete, create as _create, attach, detach

RESOURCE = "iam/groups"


def list(session: Session = None, **filters):
    """
    List a array of groups

    Args:
      session (Session): Used only if want to use a different session instead default
      filters (Any): filters to pass to filter the list of groups

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> list(page=1,page_size=10,name="group name")
    """
    result = _list(resource=RESOURCE, session=session, **filters)
    return result


def retrieve(uuid: str, session: Session = None):
    """
    Retrieve a group by uuid

    Args:
      uuid (string): group uuid to retrieve
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> retrieve("f8b85a7a-4540-4d46-a2ed-00e6134ee84a")
    """
    result = _retrieve(RESOURCE, id=uuid, session=session)
    return result


def update(uuid: str, model: Dict, session: Session = None):
    """
    Update a group by passing uuid and an model (Data to update)

    Args:
      uuid (string): group uuid to retrieve
      model (Dict): A dictionary with the data the will be updated on group
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> update("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", { "name": "new name" })
    """
    result = _update(RESOURCE, id=uuid, payload=model, session=session)
    return result


def create(model: Dict, session: Session = None):
    """
    Create a group by passing an model (Data)

    Args:
      model (Dict): A dictionary with the data the will be used to create an group
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json
    """
    result = _create(RESOURCE, payload=model, session=session)
    return result


def delete(uuid: str, session: Session = None):
    """
    Delete a group by passing uuid

    Args:
      uuid (string): group uuid to delete
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> delete("f8b85a7a-4540-4d46-a2ed-00e6134ee84a")
    """
    result = _delete(RESOURCE, id=uuid, session=session)
    return result


def attachUser(user: str, group: str, session: Session = None):
    """
    Attach a user to a group

    Args:
      user (string): user uuid to attach on group
      group (string): group uuid that the user is going to be attached
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> attachUser("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", f3a4d78a-4540-4d46-a2ed-00e613a65d4a)
    """
    result = attach(RESOURCE, type="user", this=user,
                    that=group, session=session)
    return result


def detachUser(user: str, group: str, session: Session = None):
    """
    Detach a user from a group

    Args:
      user (string): user uuid to detach from group
      group (string): group uuid that the user is, to be detached of
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> detachUser("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", f3a4d78a-4540-4d46-a2ed-00e613a65d4a)
    """
    result = detach(RESOURCE, type="user", this=user,
                    that=group, session=session)
    return result


def attachPolicy(policy: str, group: str, session: Session = None):
    """
    Attach a policy to a group

    Args:
      policy (string): policy uuid to attach on group
      group (string): group uuid that the policy is going to be attached
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> attachPolicy("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", f3a4d78a-4540-4d46-a2ed-00e613a65d4a)
    """
    result = attach(RESOURCE, type="policy", this=policy,
                    that=group, session=session)
    return result


def detachPolicy(policy: str, group: str, session: Session = None):
    """
    Detach a policy from a group

    Args:
      policy (string): policy uuid to detach from group
      group (string): group uuid that the policy is, to be detached of
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> detachPolicy("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", f3a4d78a-4540-4d46-a2ed-00e613a65d4a)
    """
    result = detach(RESOURCE, type="policy", this=policy,
                    that=group, session=session)
    return result



from typing import Dict

from sara_sdk.common.session import Session
from ...utils.rest import retrieve as _retrieve, list_paginated as _list_paginated, list as _list, update as _update, delete as _delete, create as _create

RESOURCE = "missions/steps"


def list(session: Session = None, **filters):
    """
    List a array of steps

    Args:
      session (Session): Used only if want to use a different session instead default
      filters (Any): filters to pass to filter the list of steps

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> list(page=1,page_size=10,name="step name")
    """
    result = _list(resource=RESOURCE, session=session, **filters)
    return result


def list_paginated(session: Session = None, **filters):
    """
    List iterator of pages of steps

    Args:
      session (Session): Used only if want to use a different session instead default
      filters (Any): filters to pass to filter the list of steps

    Returns:
      result (json): returns the result of the request as json by page

    Example:
      >>> next(list(page=1,page_size=10,name="step name"))
    """
    result = _list_paginated(resource=RESOURCE, session=session,
                             version="v2", **filters)
    return result


def retrieve(uuid: str, session: Session = None):
    """
    Retrieve a step by uuid

    Args:
      uuid (string): step uuid to retrieve
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> retrieve("f8b85a7a-4540-4d46-a2ed-00e6134ee84a")
    """
    result = _retrieve(RESOURCE, id=uuid, session=session, version="v2")
    return result


def update(uuid: str, model: Dict, session: Session = None):
    """
    Update a step by passing uuid and an model (Data to update)

    (Only super step can do it!)

    Args:
      uuid (string): step uuid to retrieve
      model (Dict): A dictionary with the data the will be updated on step
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> update("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", { "name": "new name" })
    """
    result = _update(RESOURCE, id=uuid, payload=model,
                     session=session,  version="v2")
    return result


def create(model: Dict, session: Session = None):
    """
    Create a step by passing an model (Data)

    (Only super step can do it!)

    Args:
      model (Dict): A dictionary with the data the will be used to create an step
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json
    """
    result = _create(RESOURCE, payload=model, session=session, version="v2")
    return result


def delete(uuid: str, session: Session = None):
    """
    Delete a step by passing uuid

    (Only super step can do it!)

    Args:
      uuid (string): step uuid to delete
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> delete("f8b85a7a-4540-4d46-a2ed-00e6134ee84a")
    """
    result = _delete(RESOURCE, id=uuid, session=session, version="v2")
    return result



class SaraSdkError(Exception):
    pass


class Error(SaraSdkError):

    def __init__(self, code, message):
        super(Exception, self).__init__(
            "{code}: {message}".format(code=code, message=message))
        self.code = code
        self.message = message


class InternalServerError(SaraSdkError):

    def __init__(self, message="Internal Server Error"):
        super(Exception, self).__init__(message)


class UnknownError(SaraSdkError):

    def __init__(self, message):
        super(Exception, self).__init__(
            "Unkown exception encountered: {}".format(message))


class AuthorizationError(SaraSdkError):

    def __init__(self, message="Session not auth, try using sara_sdk.auth(access_key, secret_key) before calling functions"):
        super(Exception, self).__init__(message)


class InvalidSignatureError(SaraSdkError):
    pass



from dotenv import load_dotenv
from .common.session import Session
from .core import iam, missions


__version__ = "1.0.0"
__timeout__ = 15

load_dotenv()

# TODO: set core folder as default path


DEFAULT_SESSION = None
API_URL = "https://sara.synkar.com/"
AUTH_URL = "https://auth.sara.synkar.com/oauth2/token"


def setup_default_session(access_key, secret_access_key, scope=None):
    """
    Set up a default session, using the access_key and secret_key and call auth to authenticate

    Args:
      access (string): Access Key used to authenticate the user on Sara API
      secret (string): Secret Key used to authenticate the user on Sara API

    """
    global DEFAULT_SESSION
    DEFAULT_SESSION = Session(access_key=access_key,
                              secret_access_key=secret_access_key, scope=scope)
    DEFAULT_SESSION.auth()


def auth(access, secret, scope=None):
    """
    Check if a default_session is defined if not call the setup_default_session to create a new session and auth
    If is already defined only call auth again to update keys

    Args:
      access (string): Access Key used to authenticate the user on Sara API
      secret (string): Secret Key used to authenticate the user on Sara API

    Returns:
      Session: return a object of the session class

    Examples:
      >>> result = auth("AKIAIOSFODNN7EXAMPLE", "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY")
      >>> print(result.access_token)
        eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJleGFtcGxlIn0.wqwdbCdHDEoknvUP-kyH2r5NrP4dVdeGeYpEz77MGtc

      Obs: Not valid tokens above, for example purpose only
    """
    if DEFAULT_SESSION is None:
        setup_default_session(
            access_key=access, secret_access_key=secret, scope=scope)
    else:
        DEFAULT_SESSION.auth()

    return DEFAULT_SESSION



from json import dumps, loads
from time import time
from ..utils.url import urlencode
from ..error import Error, InternalServerError, UnknownError, AuthorizationError
from sys import version_info as python_version
import sara_sdk


class Response:
    """
    Response Class to receive data from the API

    Args:
        status (number): Number of the status code send by the API
        content (bytes): The content of the response in bytes

    Returns:
        Response: An instance of Response

    Example:
        >>> Response(status=200,content=bytes("content"))
    """

    def __init__(self, status, content):
        self.status = status
        self.content = content

    def json(self):
        """
        Convert the content to JSON format

        Args:
            None

        Returns:
            The content of the Response as JSON

        Example:
            >>> result = Response.json()
            >>> print(result)
                { key: "data" }
        """
        if self.content:
            return loads(self.content.decode("utf-8"))
        else:
            return {}


def fetch(method, path, payload=None, query=None, session=None, version="v1"):
    """
    Function to do requests on sara api

    Args:
        method (Request): method of request from the python-requests package
        path (string): the url path to where the request will be send
        payload (Dict): payload to send through the request
        query (string): query to insert on the end of url request
        session (Session): session where to get the authorization (only needed if doens't want
        to use the DEFAULT_SESSION)
        version (string): the version of the api

    Returns:
        Response: the response of the API

    Examples:
        >>> fetch(method=post, path="iam/robots")
    """
    url = sara_sdk.API_URL + version

    if query is not None:
        url = "{base_url}/{path}/{query}".format(
            base_url=url, path=path, query=urlencode(query))
    else:
        url = "{base_url}/{path}/".format(
            base_url=url, path=path)

    agent = "Python-{major}.{minor}.{micro}-SDK-{sdk_version}".format(
        major=python_version.major,
        minor=python_version.minor,
        micro=python_version.micro,
        sdk_version=sara_sdk.__version__,
    )

    if session is None:
        session = sara_sdk.DEFAULT_SESSION

    access_time = time()

    # if session has expired get a new access_token
    if access_time >= session.expires_in:
        session.auth()

    body = payload
    bearer_token = "Bearer {token}".format(token=session.access_token)

    try:
        request = method(
            url=url,
            data=body,
            headers={
                "Access-Time": str(access_time),
                "Content-Type": "application/x-www-form-urlencoded",
                "User-Agent": agent,
                "Accept-Language": "en-US",
                "Authorization": bearer_token
            },
            timeout=sara_sdk.__timeout__
        )
    except Exception as exception:
        raise UnknownError("{}: {}".format(
            exception.__class_.__name__, str(exception)))

    response = Response(status=request.status_code, content=request.content)

    if response.status == 500:
        raise InternalServerError()
    if response.status == 400:
        raise Error(response.status, response.json()["detail"])
    if response.status == 401:
        if(session.attemps == 0):
            session.attemps += 1
            session.auth()
            fetch(method=method, path=path, payload=payload,
                  query=query, session=session, version=version)
        else:
            raise AuthorizationError()
    if response.status >= 300:
        raise UnknownError(response.content)

    return response



from typing import Dict

from sara_sdk.common.session import Session
from ...utils.rest import retrieve as _retrieve, list as _list, update as _update, delete as _delete, create as _create

RESOURCE = "iam/services"


def list(session: Session = None, **filters):
    """
    List a array of services

    Args:
      session (Session): Used only if want to use a different session instead default
      filters (Any): filters to pass to filter the list of services

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> list(page=1,page_size=10,name="service name")
    """
    result = _list(resource=RESOURCE, session=session, **filters)
    return result


def retrieve(uuid: str, session: Session = None):
    """
    Retrieve a service by uuid

    Args:
      uuid (string): service uuid to retrieve
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> retrieve("f8b85a7a-4540-4d46-a2ed-00e6134ee84a")
    """
    result = _retrieve(RESOURCE, id=uuid, session=session)
    return result


def create(model: Dict, session: Session = None):
    """
    Create a service by passing an model (Data)

    Args:
      model (Dict): A dictionary with the data the will be used to create an service
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json
    """
    result = _create(RESOURCE, payload=model, session=session)
    return result



from typing import Dict
import json
from sara_sdk.common.session import Session
from ...utils.rest import list as _list, list_paginated as _list_paginated, create as _create

RESOURCE = "missions"


def list(session: Session = None, **filters):
    """
    List a array of missions

    Args:
      session (Session): Used only if want to use a different session instead default
      filters (Any): filters to pass to filter the list of missions

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> list(page=1,page_size=10,name="mission name")
    """
    result = _list(resource=RESOURCE, session=session, **filters)
    return result


def list_paginated(session: Session = None, **filters):
    """
    List iterator of pages of missions

    Args:
      session (Session): Used only if want to use a different session instead default
      filters (Any): filters to pass to filter the list of missions

    Returns:
      result (json): returns the result of the request as json by page

    Example:
      >>> next(list(page=1,page_size=10,name="mission name"))
    """
    result = _list_paginated(resource=RESOURCE, session=session,
                             version="v2", **filters)
    return result


def last(robot: str, session: Session = None):
    """
    Retrieve the last mission by robot id

    Args:
      robot (UUID): robot to return a mission
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> retrieve("f8b85a7a-4540-4d46-a2ed-00e6134ee84a")
    """
    result = _list(RESOURCE+"/last", session=session,
                   version="v2", robot_id=robot)
    return result


def create(robot: str, stages: Dict, session: Session = None):
    """
    Create a mission by passing an model (Data)

    Args:
      robot (UUID): robot uuid to create mission
      stages (Dict): A dictionary with the data the will be used to create an mission
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json
    """
    model = {
        "robot": robot,
        "stages": json.dumps(stages)
    }
    result = _create(RESOURCE, payload=model, session=session, version="v2")
    return result



from sara_sdk.common.session import Session
from ...utils.rest import retrieve as _retrieve, list as _list, update as _update, delete as _delete, create as _create, attach, detach

RESOURCE = "iam/clients"


def list(session: Session = None, **filters):
    """
    List a array of clients

    Args:
      session (Session): Used only if want to use a different session instead default
      filters (Any): filters to pass to filter the list of clients

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> list(page=1,page_size=10,name="client name")
    """
    result = _list(resource=RESOURCE, session=session, **filters)
    return result


def attachRobot(robot: str, client: str, session: Session = None):
    """
    Attach a robot to a client

    Args:
      robot (string): robot uuid to attach on client
      client (string): client uuid that the robot is going to be attached
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> attachRobot("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", f3a4d78a-4540-4d46-a2ed-00e613a65d4a)
    """
    result = attach(RESOURCE, type="robot", this=robot,
                    that=client, session=session)
    return result


def detachRobot(robot: str, client: str, session: Session = None):
    """
    Detach a robot from a client

    Args:
      robot (string): robot uuid to detach from client
      client (string): client uuid that the robot is, to be detached of
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> detachRobot("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", f3a4d78a-4540-4d46-a2ed-00e613a65d4a)
    """
    result = detach(RESOURCE, type="robot", this=robot,
                    that=client, session=session)
    return result


def attachUser(user: str, client: str, session: Session = None):
    """
    Attach a user to a client

    Args:
      user (string): user uuid to attach on client
      client (string): client uuid that the user is going to be attached
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> attachUser("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", f3a4d78a-4540-4d46-a2ed-00e613a65d4a)
    """
    result = attach(RESOURCE, type="user", this=user,
                    that=client, session=session)
    return result


def detachUser(user: str, client: str, session: Session = None):
    """
    Detach a user from a client

    Args:
      user (string): user uuid to detach from client
      client (string): client uuid that the user is, to be detached of
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> detachUser("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", f3a4d78a-4540-4d46-a2ed-00e613a65d4a)
    """
    result = detach(RESOURCE, type="user", this=user,
                    that=client, session=session)
    return result


def slug(slug: str, session=None):
    """
    Get the data of an client by passing the slug

    Args:
      slug (str): slug of the client
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> slug("test-client")
    """
    result = _list("iam/slugs", session=session, pk=slug)
    return result



from typing import Dict

from sara_sdk.common.session import Session
from ...utils.rest import retrieve as _retrieve, list as _list, update as _update, delete as _delete, create as _create, attach, detach

RESOURCE = "iam/fleets"


def list(session: Session = None, **filters):
    """
    List a array of fleets

    Args:
      session (Session): Used only if want to use a different session instead default
      filters (Any): filters to pass to filter the list of fleets

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> list(page=1,page_size=10,name="fleetname")
    """
    result = _list(resource=RESOURCE, session=session, **filters)
    return result


def retrieve(uuid: str, session: Session = None):
    """
    Retrieve a fleet by uuid

    Args:
      uuid (string): fleet uuid to retrieve
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> retrieve("f8b85a7a-4540-4d46-a2ed-00e6134ee84a")
    """
    result = _retrieve(RESOURCE, id=uuid, session=session)
    return result


def update(uuid: str, model: Dict, session: Session = None):
    """
    Update a fleet by passing uuid and an model (Data to update)

    Args:
      uuid (string): fleet uuid to retrieve
      model (Dict): A dictionary with the data the will be updated on fleet
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> update("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", { "name": "new name" })
    """
    result = _update(RESOURCE, id=uuid, payload=model, session=session)
    return result


def create(model: Dict, session: Session = None):
    """
    Create a fleet by passing an model (Data)

    Args:
      model (Dict): A dictionary with the data the will be used to create an fleet
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json
    """
    result = _create(RESOURCE, payload=model, session=session)
    return result


def delete(uuid: str, session: Session = None):
    """
    Delete a fleet by passing uuid

    Args:
      uuid (string): fleet uuid to delete
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of the request as json

    Example:
      >>> delete("f8b85a7a-4540-4d46-a2ed-00e6134ee84a")
    """
    result = _delete(RESOURCE, id=uuid, session=session)
    return result


def attachRobot(robot: str, fleet: str, session: Session = None):
    """
    Attach a robot to a fleet

    Args:
      robot (string): robot uuid to attach on fleet
      fleet (string): fleet uuid that the robot is going to be attached
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> attachRobot("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", f3a4d78a-4540-4d46-a2ed-00e613a65d4a)
    """
    result = attach(RESOURCE, type="robot", this=robot,
                    that=fleet, session=session)
    return result


def detachRobot(robot: str, fleet: str, session: Session = None):
    """
    Detach a robot from a fleet

    Args:
      robot (string): robot uuid to detach from fleet
      fleet (string): fleet uuid that the robot is, to be detached of
      session (Session): Used only if want to use a different session instead default

    Returns:
      result (json): returns the result of request as json

    Example:
      >>> detachRobot("f8b85a7a-4540-4d46-a2ed-00e6134ee84a", f3a4d78a-4540-4d46-a2ed-00e613a65d4a)
    """
    result = detach(RESOURCE, type="robot", this=robot,
                    that=fleet, session=session)
    return result



from urllib.parse import urlencode as _urlencode


def urlencode(params):
    params = {k: valueToString(v) for k, v in params.items()}
    return "?" + _urlencode(params) if params else ""


def valueToString(value):
    if isinstance(value, (tuple, list, set)):
        return ",".join([valueToString(value) for value in value])
    return str(value)
